program inpainting_test
    
    use alm_tools
    use fitstools
    use healpix_types
    use pix_tools
    use udgrade_nr
    use rngmod
    implicit none
    
    integer :: funit_one, funit_two, funit_three, funit_four, &
               funit_five, funit_six, funit_seven, n_side_recon, &
               n_pix_recon, n_dims = 1, l_max_recon, l_max_out, &
               m_max_out, i, k, ell, em, n_pix_unmasked, &
               n_real_a_lms, n_cmb, n_its, seed, i_r, &
               l_min_map, l_max_map
    real(kind = dp), dimension(:, :), allocatable :: map_in, &
         map_out, mask, ring_weight, c_l
    real(kind = dp), dimension(:), allocatable :: a_lm_out_r, &
         a_lm_half_r, delta_r
    real(kind = dp), dimension(1: 2) :: bounds = (/ -1.0_dp, &
                                                     1.0_dp /)
    real(kind = dp) :: alpha, beta, cut_theta_min, cut_theta_max, &
                       cut_width, theta, phi, conv_thresh, a_lm_l_2, &
                       delta_l_2, f_sky, cut_sky_mean
    complex(kind = dpc), dimension(:, :, :), allocatable :: a_lm_in, &
         a_lm_out, a_lm_half, delta, a_lm_cut
    character(len = 80), dimension(1:10, 1) :: header
    character(len = 256) :: mask_file, c_l_file
    character(len = 10) :: k_char, n_its_char
    character(len = 6) :: mask_order
    character(len = 5) :: f_sky_char
    character(len = 3) :: n_l_char
    logical :: converged, save_all_maps, sep_inpainting, &
               seed_chosen, l_min_map_chosen, l_max_map_chosen, &
               sub_mean, norm_beta, exact_const
    type(planck_rng) :: rng_handle
    
    
    ! default parameter values
    n_side_recon = 32
    l_max_recon = 3
    l_max_out = 10
    
    
    ! CMB / mask parameters
    n_cmb = 10000
    cut_width = 60.0_dp
    mask_file = ''
    mask_order = 'ring'
    seed = 123
    seed_chosen = .false.
    c_l_file = '/Users/stephen/Data/c_l_data/' // &
               'high_l_wmap7_for_planck/camb_47267783_cls.fits'
    save_all_maps = .false.
    l_min_map_chosen = .false.
    l_max_map_chosen = .false.
    
    
    ! sparse reconstruction parameters
    sep_inpainting = .false.
    alpha = 1.0_dp
    beta = 1.0_dp
    n_its = 150
    conv_thresh = 0.001_dp
    sub_mean = .false.
    norm_beta = .false.
    exact_const = .true.
    
    
    ! read parameters from command-line
    call parse_options()
    write(n_l_char, '(I3.1)') l_max_out - 1
    
    
    ! calculate pixel counts for input, a_lm-reconstruction and output
    ! map resolutions
    n_pix_recon = nside2npix(n_side_recon)
    
    
    ! define limits in m to be limits in m
    m_max_out = l_max_out
    
    
    ! calculate the number of real a_lms needed if separate inpainting
    ! is being used
    n_real_a_lms = (l_max_out + 1) ** 2
    
    
    ! determine the multipoles to include in the maps (if saving them)
    if (.not. l_min_map_chosen) l_min_map = 2
    if (.not. l_max_map_chosen) l_max_map = 3
    
    
    ! allocate standard required arrays
    allocate(map_in(0: n_pix_recon - 1, n_dims))
    allocate(map_out(0: n_pix_recon - 1, n_dims))
    allocate(mask(0: n_pix_recon - 1, n_dims))
    allocate(c_l(0: l_max_out, 1))
    allocate(a_lm_in(1, 0: l_max_out, 0: m_max_out))
    allocate(a_lm_out(1, 0: l_max_out, 0: m_max_out))
    allocate(a_lm_half(1, 0: l_max_out, 0: m_max_out))
    allocate(a_lm_cut(1, 0: l_max_out, 0: m_max_out))
    if (sep_inpainting) then
       
       allocate(a_lm_out_r(0: n_real_a_lms - 1))
       allocate(a_lm_half_r(0: n_real_a_lms - 1))
       allocate(delta_r(0: n_real_a_lms - 1))
       
    else
       
       allocate(delta(1, 0: l_max_out, 0: m_max_out))
       
    end if
    allocate(ring_weight(1: 2 * n_side_recon, 1))
    ring_weight = 1.0_dp
    
    
    ! either read in a user-specified mask or create a large Galactic
    ! mask: for simplicity azimuthally symmetric
    if (trim(adjustl(mask_file)) .ne. '') then
       
       call input_map(trim(adjustl(mask_file)), mask, n_pix_recon, 1)
       if (trim(adjustl(mask_order)) .ne. 'ring') then
          
          call convert_nest2ring(n_side_recon, mask)
          
       end if
       
    else
       
       mask(:, 1) = 1.0_dp
       if (cut_width .ne. 0.0_dp) then
          
          cut_theta_min = (90.0_dp - cut_width / 2.0_dp) * &
                          pi / 180.0_dp
          cut_theta_max = (90.0_dp + cut_width / 2.0_dp) * &
                          pi / 180.0_dp
          do k = 0, n_pix_recon - 1
             
             call pix2ang_ring(n_side_recon, k, theta, phi)
             if (theta .ge. cut_theta_min .and. &
                 theta .le. cut_theta_max) then
                
                mask(k, 1) = 0.0_dp
                
             end if
             
          end do
          
       end if
       
    end if
    f_sky = sum(mask(:, 1)) / real(n_pix_recon, kind = dp)
    write(f_sky_char, '(F5.3)') f_sky
    print*, 'f_sky unmasked: ' // f_sky_char
    
    
    ! read in power spectrum and rescale to uK
    call fits2cl(c_l_file, c_l, l_max_out, 1, header(:, 1))
    c_l = c_l * 1.0D12
    
    
    ! check if the user has provided a seed, and generate one (and an
    ! associated rng_handle) if not. initialise the random seed in 
    ! line with the HEALPix SYNFAST procedure (see 
    ! $HEALPIX/src/f90/synfast/syn_sub_inc.f90): "Seed should 
    ! initially be set to a large, odd integer"
    if (.not. seed_chosen) then
       
       call system_clock(count = seed)
       
    end if
    if (mod(seed, 2) .eq. 0) seed = seed + 1
    call rand_init(rng_handle, seed)
    
    
    ! optionally match MRS's adjustment to beta for direct code 
    ! comparison
    if (norm_beta) then
       
       beta = beta * sqrt(4.0_dp * pi / real(n_pix_recon, kind = dp))
       
    end if
    
    
    ! open up output files
    funit_one = 666
    funit_two = 667
    funit_three = 668
    funit_four = 669
    funit_five = 670
    funit_six = 671
    funit_seven = 672
    open (unit = funit_one, file = 'input_a_2m.txt', &
          status = 'replace')
    open (unit = funit_two, file = 'input_a_3m.txt', &
          status = 'replace')
    if (sep_inpainting) then
       
       open (unit = funit_three, file = 'sep_inpainted_a_2m.txt', &
             status = 'replace')
       open (unit = funit_four, file = 'sep_inpainted_a_3m.txt', &
             status = 'replace')
       open (unit = funit_seven, file = 'sep_inpainted_a_l0.txt', &
             status = 'replace')
       
    else
       
       open (unit = funit_three, file = 'inpainted_a_2m.txt', &
             status = 'replace')
       open (unit = funit_four, file = 'inpainted_a_3m.txt', &
             status = 'replace')
       open (unit = funit_seven, file = 'inpainted_a_l0.txt', &
             status = 'replace')
       
    end if
    open (unit = funit_five, file = 'cs_a_2m.txt', &
          status = 'replace')
    open (unit = funit_six, file = 'cs_a_3m.txt', &
          status = 'replace')
    header(:, 1) = ''
    
    
    ! generate a set of full-sky Gaussian CMB realisations
    k = 0
    do while (k .lt. n_cmb)
       
       
       ! generate random map
       write(k_char, '(I10.1)') k + 1
       print*, 'simulation: ' // trim(adjustl(k_char))
       call c_l_to_a_lm(c_l(:, 1), a_lm_in, rng_handle)
       call alm2map(n_side_recon, l_max_out, m_max_out, a_lm_in, &
                    map_in(:, 1))
       
       
       ! generate the cut-sky a_lms, which will form our first guess 
       ! at the inpainted a_lms. optionally subtract the mean of the 
       ! cut-sky map from the unmasked pixels in the input map. we use
       ! the input map to constrain the reconstruction, so subtracting
       ! the mean here ensures we constrain using the zero-mean cut-
       ! sky map every time. note that the input a_lms are not changed
       ! by this; since we use these to build the map we eventually 
       ! save to file (if requested), this means the true input map is
       ! saved, not the version with the cut-sky mean subtracted
       if (sub_mean) then
          
          cut_sky_mean = sum(map_in(:, 1) * mask(:, 1)) / &
                         sum(mask(:, 1))
          where (mask(:, 1) .eq. 1.0_dp)
             
             map_in(:, 1) = map_in(:, 1) - cut_sky_mean
             
          end where
          
       end if
       map_out(:, 1) = map_in(:, 1) * mask(:, 1)
       call map2alm(n_side_recon, l_max_out, m_max_out, map_out(:, 1), &
                    a_lm_out, bounds, ring_weight)
       a_lm_cut = a_lm_out
       
       
       ! choose between two sparse-inpainting techniques
       if (sep_inpainting) then
          
          
          ! apply a sparsity prior to the real and imaginary parts of
          ! the a_lms. this is equivalent to applying a sparsity prior
          ! to the real a_lms
          do i = 0, n_its - 1
             
             
             ! + 1/2 step: convert estimated a_lms to real space and 
             ! constrain with the observed pixel values (projection step).
             ! NB: the (:, 1) in map_out is critical here
             call alm2map(n_side_recon, l_max_out, l_max_out, &
                          a_lm_out, map_out(:, 1))
             where (mask(:, 1) .eq. 1.0_dp)
                
                map_out(:, 1) = map_in(:, 1)
                
             end where
             call map2alm(n_side_recon, l_max_out, m_max_out, &
                          map_out(:, 1), a_lm_half, bounds, &
                          ring_weight)
             
             
             ! loop through a_lms to update, calculating the two-norms
             ! of the a_lms and the updates to perform a convergence
             ! check
             a_lm_l_2 = 0.0_dp
             delta_l_2 = 0.0_dp
             call a_lm_complex_to_real(a_lm_out, a_lm_out_r)
             call a_lm_complex_to_real(a_lm_half, a_lm_half_r)
             delta_r = (2.0_dp * a_lm_half_r) - a_lm_out_r
             do ell = 0, l_max_out
                
                do em = -ell, ell
                   
                   i_r = ell ** 2 + ell + em! + 1 - 1
                   if (abs(a_lm_out_r(i_r)) .eq. 0.0_dp .and. &
                       f_sky .eq. 1.0_dp) then
                      
                      
                      ! this code is only called when using full-sky
                      ! data!
                      ! if |a_lm_out| = 0, we should stop iterating. 
                      ! all we need to do in this case is to add to 
                      ! the 2-norm of the a_lms. stopping iterating 
                      ! ensures that the a_lm_out in the case of full-
                      ! sky data is exact soft-thresholding of the 
                      ! a_lm_in (the a_lm_half are always set to 
                      ! a_lm_in). this portion of code makes no 
                      ! difference to masked data
                      a_lm_l_2 = a_lm_l_2 + a_lm_out_r(i_r) ** 2
                      
                      
                   else
                      
                      
                      ! + 1 step: use the a_lm_half to adjust the 
                      ! previous iteration's a_lms according to the 
                      ! threshold, beta (proximity step)
                      if (abs(delta_r(i_r)) .gt. beta) then
                         
                         delta_r(i_r) = delta_r(i_r) / &
                                        abs(delta_r(i_r)) * &
                                        (abs(delta_r(i_r)) - beta)
                         
                      else
                         
                         delta_r(i_r) = 0.0_dp
                         
                      end if
                      
                      
                      ! calculate two-norms of a_lms and update
                      a_lm_l_2 = a_lm_l_2 + a_lm_out_r(i_r) ** 2
                      delta_l_2 = delta_l_2 + &
                                  (alpha * (delta_r(i_r) - &
                                  a_lm_half_r(i_r))) ** 2
                      
                      
                      ! update a_lms
                      a_lm_out_r(i_r) = a_lm_out_r(i_r) + alpha * &
                                        (delta_r(i_r) - &
                                         a_lm_half_r(i_r))
                      
                      
                   end if
                   
                end do
                
             end do
             
             
             ! check for convergence (and convert back to complex 
             ! a_lms)
             call a_lm_real_to_complex(a_lm_out_r, a_lm_out)
             converged = .false.
             if (sqrt(delta_l_2 / a_lm_l_2) .le. conv_thresh) then
                
                write(n_its_char, '(I10.1)') i + 1
                print*, 'converged in ' // &
                        trim(adjustl(n_its_char)) // &
                        ' iterations'
                converged = .true.
                exit
                
             end if
             
             
          end do
          
          
       else
          
          
          ! apply a sparsity prior to the magnitude of the complex 
          ! a_lms
          do i = 0, n_its - 1
             
             
             ! + 1/2 step: convert estimated a_lms to real space and 
             ! constrain with the observed pixel values (projection 
             ! step). NB: the (:, 1) in map_out is critical here
             call alm2map(n_side_recon, l_max_out, l_max_out, &
                          a_lm_out, map_out(:, 1))
             where (mask(:, 1) .eq. 1.0_dp)
                
                map_out(:, 1) = map_in(:, 1)
                
             end where
             call map2alm(n_side_recon, l_max_out, m_max_out, &
                          map_out(:, 1), a_lm_half, bounds, &
                          ring_weight)
             
             
             ! loop through a_lms to update, calculating the two-norms
             ! of the a_lms and the updates to perform a convergence
             ! check
             a_lm_l_2 = 0.0_dp
             delta_l_2 = 0.0_dp
             delta = (2.0_dp * a_lm_half) - a_lm_out
             do ell = 0, l_max_out
                
                do em = 0, ell
                   
                   if (abs(a_lm_out(1, ell, em)) .eq. 0.0_dp .and. &
                       f_sky .eq. 1.0_dp) then
                      
                      
                      ! this code is only called when using full-sky
                      ! data!
                      ! if |a_lm_out| = 0, we should stop iterating. 
                      ! all we need to do in this case is to add to 
                      ! the 2-norm of the a_lms. stopping iterating 
                      ! ensures that the a_lm_out in the case of full-
                      ! sky data is exact soft-thresholding of the 
                      ! a_lm_in (the a_lm_half are always set to 
                      ! a_lm_in). this portion of code makes no 
                      ! difference to masked data
                      if (em .eq. 0) then
                         
                         a_lm_l_2 = a_lm_l_2 + &
                                    abs(a_lm_out(1, ell, 0)) ** 2
                         
                      else
                         
                         a_lm_l_2 = a_lm_l_2 + 2.0_dp * &
                                    abs(a_lm_out(1, ell, em)) ** 2
                         
                      end if
                      
                      
                   else
                      
                      
                      ! otherwise...
                      ! + 1 step: use the a_lm_half to adjust the 
                      ! previous iteration's a_lms according to the 
                      ! threshold, beta (the proximity step)
                      if (abs(delta(1, ell, em)) .gt. beta) then
                         
                         delta(1, ell, em) = delta(1, ell, em) / &
                                             abs(delta(1, ell, em)) * &
                                             (abs(delta(1, ell, em)) - &
                                              beta)
                         
                      else
                         
                         delta(1, ell, em) = cmplx(0.0, 0.0, kind = dpc)
                         
                      end if
                      
                      
                      ! then calculate the two-norms of a_lms and the 
                      ! updates to them
                      if (em .eq. 0) then
                         
                         a_lm_l_2 = a_lm_l_2 + &
                                    abs(a_lm_out(1, ell, 0)) ** 2
                         delta_l_2 = delta_l_2 + &
                                     abs(alpha * (delta(1, ell, 0) - &
                                         a_lm_half(1, ell, 0))) ** 2
                         
                      else
                         
                         a_lm_l_2 = a_lm_l_2 + 2.0_dp * &
                                    abs(a_lm_out(1, ell, em)) ** 2
                         delta_l_2 = delta_l_2 + 2.0_dp * &
                                     abs(alpha * (delta(1, ell, em) - &
                                         a_lm_half(1, ell, em))) ** 2
                         
                      end if
                      
                      
                      ! and finally update
                      a_lm_out(1, ell, em) = a_lm_out(1, ell, em) + &
                                             alpha * (delta(1, ell, em) - &
                                                      a_lm_half(1, ell, em))
                      
                      
                   end if
                   
                   
                end do
                
             end do
             
             
             ! check for convergence
             converged = .false.
             if (sqrt(delta_l_2 / a_lm_l_2) .le. conv_thresh) then
                
                write(n_its_char, '(I10.1)') i + 1
                print*, 'converged in ' // &
                        trim(adjustl(n_its_char)) // ' iterations'
                converged = .true.
                exit
                
             end if
             
             
          end do
          
          
       end if
       
       
       ! if routine did not converge, re-draw the a_lms
       if (converged) then
          
          k = k + 1
          
       else
          
          print*, 'WARNING: convex optimisation not converged!'
          cycle
          
       end if
       
       
       ! the Douglas-Rachford algorithm consists of two steps. the 
       ! a_lms produced by the first ("+ 1/2") step (called a_lm_half) 
       ! are the global minimizer of the sparse inpainting problem; 
       ! i.e. what most users want. the a_lms produced by the second 
       ! ("+ 1") step (currently called a_lm_out) can also be returned
       ! if the user requests (this is currently the case, for 
       ! example, in the MRS code). from this point onwards, the 
       ! chosen a_lms are renamed to a_lm_out
       if (exact_const) then
          
          a_lm_out = a_lm_half
          
       end if
       
       
       ! output the l = {2, 3} input a_lms to file for processing 
       ! with IDL
       write(funit_one, fmt = '(6(E19.12, 1X))'), &
             real(a_lm_in(1, 2, 0)), &
             aimag(a_lm_in(1, 2, 0)), &
             real(a_lm_in(1, 2, 1)), &
             aimag(a_lm_in(1, 2, 1)), &
             real(a_lm_in(1, 2, 2)), &
             aimag(a_lm_in(1, 2, 2))
       write(funit_two, fmt = '(8(E19.12, 1X))'), &
             real(a_lm_in(1, 3, 0), kind = dp), &
             aimag(a_lm_in(1, 3, 0)), &
             real(a_lm_in(1, 3, 1), kind = dp), &
             aimag(a_lm_in(1, 3, 1)), &
             real(a_lm_in(1, 3, 2), kind = dp), &
             aimag(a_lm_in(1, 3, 2)), &
             real(a_lm_in(1, 3, 3), kind = dp), &
             aimag(a_lm_in(1, 3, 3))
       
       
       ! output the l = {2, 3} cut-sky a_lms to file
       write(funit_five, fmt = '(6(E19.12, 1X))'), &
             real(a_lm_cut(1, 2, 0)), &
             aimag(a_lm_cut(1, 2, 0)), &
             real(a_lm_cut(1, 2, 1)), &
             aimag(a_lm_cut(1, 2, 1)), &
             real(a_lm_cut(1, 2, 2)), &
             aimag(a_lm_cut(1, 2, 2))
       write(funit_six, fmt = '(8(E19.12, 1X))'), &
             real(a_lm_cut(1, 3, 0), kind = dp), &
             aimag(a_lm_cut(1, 3, 0)), &
             real(a_lm_cut(1, 3, 1), kind = dp), &
             aimag(a_lm_cut(1, 3, 1)), &
             real(a_lm_cut(1, 3, 2), kind = dp), &
             aimag(a_lm_cut(1, 3, 2)), &
             real(a_lm_cut(1, 3, 3), kind = dp), &
             aimag(a_lm_cut(1, 3, 3))
       
       
       ! output map if wished, using only selected multipoles 
       ! (default being ell = {2, 3}
       if (save_all_maps) then
          
          do ell = 0, l_min_map - 1
             
             a_lm_out(1, ell, :) = 0.0_dp
             
          end do
          do ell = l_max_map + 1, l_max_out
             
             a_lm_out(1, ell, :) = 0.0_dp
             
          end do
          call alm2map(n_side_recon, l_max_out, l_max_out, a_lm_out, &
                       map_out(:, 1))
          call output_map(map_out, header(:, 1), &
                          '!inpainted_map_out.fits')
          
       end if
       
       
       ! also output the l = {2, 3} l-1 inpainted a_lms to file
       write(funit_three, fmt = '(6(E19.12, 1X))'), &
             real(a_lm_out(1, 2, 0)), &
             aimag(a_lm_out(1, 2, 0)), &
             real(a_lm_out(1, 2, 1)), &
             aimag(a_lm_out(1, 2, 1)), &
             real(a_lm_out(1, 2, 2)), &
             aimag(a_lm_out(1, 2, 2))
       write(funit_four, fmt = '(8(E19.12, 1X))'), &
             real(a_lm_out(1, 3, 0), kind = dp), &
             aimag(a_lm_out(1, 3, 0)), &
             real(a_lm_out(1, 3, 1), kind = dp), &
             aimag(a_lm_out(1, 3, 1)), &
             real(a_lm_out(1, 3, 2), kind = dp), &
             aimag(a_lm_out(1, 3, 2)), &
             real(a_lm_out(1, 3, 3), kind = dp), &
             aimag(a_lm_out(1, 3, 3))
       
       
       ! also output the l = {2: l_max}, m = 0 l-1-inpainted a_lms to
       ! file
       write(funit_seven, fmt = '(' // trim(adjustl(n_l_char)) // &
                               '(E19.12, 1X))'), &
             real(a_lm_out(1, 2: l_max_out, 0))
       
       
       ! output input map if wished
       if (save_all_maps) then
          
          do ell = 0, l_min_map - 1
             
             a_lm_in(1, ell, :) = 0.0_dp
             
          end do
          do ell = l_max_map + 1, l_max_out
             
             a_lm_in(1, ell, :) = 0.0_dp
             
          end do
          call alm2map(n_side_recon, l_max_out, l_max_out, a_lm_in, &
                       map_out(:, 1))
          call output_map(map_out, header(:, 1), &
                          '!inpainted_map_in.fits')
          
       end if
       
       
       ! output cut-sky map if wished
       if (save_all_maps) then
          
          do ell = 0, l_min_map - 1
             
             a_lm_cut(1, ell, :) = 0.0_dp
             
          end do
          do ell = l_max_map + 1, l_max_out
             
             a_lm_cut(1, ell, :) = 0.0_dp
             
          end do
          call alm2map(n_side_recon, l_max_out, l_max_out, a_lm_cut, &
                       map_out(:, 1))
          call output_map(map_out, header(:, 1), &
                          '!inpainted_map_cut.fits')
          
       end if
       
       
    end do
    
    
    ! close files
    close (funit_one)
    close (funit_two)
    close (funit_three)
    close (funit_four)
    close (funit_five)
    close (funit_six)
    
    
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    
    
  contains
    
    
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    
    
    ! code to parse command-line options (based on HEALPix)
    subroutine parse_options()
      
      use extension, only: getArgument, nArguments
      
      implicit none
      
      integer :: n, i
      character(len = 256) :: opt
      character(len = 256) :: arg
      
      n = nArguments()
      
      do i = 1,n,2
         
         call getArgument(i,opt)
         
         if (i == n .and. trim(opt) /= '-help') then
            write(*,'(a,a,a)') 'Option ', trim(opt), ' has no argument'
            stop
         end if
         
         if(trim(opt) /= '-help') call getArgument(i+1,arg)
         
         ! Read each argument in turn
         select case (trim(opt))
            
         case ('-help')
            write(*,'(a)') 'Usage: testinpaint [-n_cmb n_cmb]'
            write(*,'(a)') '                   [-n_side n_side_recon]'
            write(*,'(a)') '                   [-cut_width cut_width]'
            write(*,'(a)') '                   [-mask_file mask_file]'
            write(*,'(a)') '                   [-mask_order mask_order]'
            write(*,'(a)') '                   [-seed seed]'
            write(*,'(a)') '                   [-c_l_file c_l_file]'
            write(*,'(a)') '                   [-l_max_recon l_max_recon]'
            write(*,'(a)') '                   [-l_max l_max_out]'
            write(*,'(a)') '                   [-sep_sparse sep_inpainting]'
            write(*,'(a)') '                   [-alpha alpha]'
            write(*,'(a)') '                   [-beta beta]'
            write(*,'(a)') '                   [-n_its n_its]'
            write(*,'(a)') '                   [-conv_thresh conv_thresh]'
            write(*,'(a)') '                   [-sub_mean sub_mean]'
            write(*,'(a)') '                   [-norm_beta norm_beta]'
            write(*,'(a)') '                   [-exact_const exact_const]'
            write(*,'(a)') '                   [-save_maps save_all_maps]'
            write(*,'(a)') '                   [-l_min_map l_min_map]'
            write(*,'(a)') '                   [-l_max_map l_max_map]'
            stop
            
            
         case ('-n_cmb')
            read(arg,*) n_cmb
            
         case ('-n_side')
            read(arg,*) n_side_recon
            
         case ('-cut_width')
            read(arg,*) cut_width
            
         case ('-mask_file')
            mask_file = trim(arg)
            
         case ('-mask_order')
            mask_order = trim(arg)
            
         case ('-seed')
            read(arg,*) seed
            seed_chosen = .true.
            
         case ('-c_l_file')
            c_l_file = trim(arg)
            
         case ('-l_max_recon')
            read(arg,*) l_max_recon
            
         case ('-l_max')
            read(arg,*) l_max_out
            
         case ('-sep_sparse')
            read(arg,*) sep_inpainting
            
         case ('-alpha')
            read(arg,*) alpha
            
         case ('-beta')
            read(arg,*) beta
            
         case ('-n_its')
            read(arg,*) n_its
            
         case ('-conv_thresh')
            read(arg,*) conv_thresh
            
         case ('-sub_mean')
            read(arg,*) sub_mean
            
         case ('-norm_beta')
            read(arg,*) norm_beta
            
         case ('-exact_const')
            read(arg,*) exact_const
            
         case ('-save_maps')
            read(arg,*) save_all_maps
            
         case ('-l_min_map')
            read(arg,*) l_min_map
            l_min_map_chosen = .true.
            
         case ('-l_max_map')
            read(arg,*) l_max_map
            l_max_map_chosen = .true.
            
         case default
            print '("Unknown option ",a," ignored")', trim(opt)            
            
         end select
         
      end do
      
    end subroutine parse_options
    
    
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    
    
    ! convert real a_lms to complex
    subroutine a_lm_real_to_complex(a_lm_real, a_lm_complex, verbose)
        
        real(kind = dp), dimension(0:), intent(in) :: a_lm_real
        complex(kind = dpc), dimension(1:, 0:, 0:), intent(inout) :: &
             a_lm_complex
        logical, optional :: verbose
        real :: time_0, time_1, time_elapsed
        integer :: n_a_lm_real, l_max, m_max, l, m
        logical :: feedback
        
        
        ! determine verbosity, a_lm counts / array size
        feedback = .false.
        if (present(verbose)) then
            
            if (verbose) feedback = .true.
            
        end if
        if (feedback) print*, 'a_lm conversion: real to complex'
        call cpu_time(time_0)
        n_a_lm_real = size(a_lm_real, 1)
        l_max = int(sqrt(real(n_a_lm_real))) - 1
        m_max = l_max
        
        
        ! convert REAL a_lms to COMPLEX:
        ! complex a_lm = cmplx(real a_l0, 0)                       if m = 0
        ! complex a_lm = cmplx(real a_lm, -(real a_l-m)) / sqrt(2) if m > 0
        ! m < 0 terms can be generated from m > 0 terms.
        ! real a_lms are stored in the (l**2 + l + m + 1)th 
        ! array element.
        a_lm_complex(1, :, :) = cmplx(0.0, 0.0, kind = dpc)
        do l = 0, l_max
            
            a_lm_complex(1, l, 0) = cmplx(a_lm_real(l**2 + l), &
                                          0.0, kind = dpc)
            do m = 1, l
                
                a_lm_complex(1, l, m) = cmplx( a_lm_real(l**2 + l + m) / sqrt(2.0_dp), &
                                              -a_lm_real(l**2 + l - m) / sqrt(2.0_dp), &
                                               kind = dpc)
                
            end do
            
        end do
        call cpu_time(time_1)
        time_elapsed = time_1 - time_0
        if (feedback) print*, 'a_lm conversion: completed in ', &
                               time_elapsed, ' secs'
        
        
    end subroutine a_lm_real_to_complex
    
    
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    
    
    ! convert complex a_lms to real
    subroutine a_lm_complex_to_real(a_lm_complex, a_lm_real, verbose)
        
        complex(kind = dpc), dimension(1:, 0:, 0:), &
             intent(inout) :: a_lm_complex
        real(kind = dp), dimension(0:), intent(inout) :: &
             a_lm_real
        logical, optional :: verbose
        real :: time_0, time_1, time_elapsed
        integer :: l_max, l, m, index
        logical :: feedback
        
        
        ! determine verbosity, a_lm counts / array size
        feedback = .false.
        if (present(verbose)) then
            
            if (verbose) feedback = .true.
            
        end if
        if (feedback) print*, 'a_lm conversion: complex to real'
        call cpu_time(time_0)
        l_max = size(a_lm_complex, 2) - 1
        
        
        ! convert COMPLEX a_lms to REAL
        ! 1) real a_lm = -sqrt(2) * IM{complex a_l-m} if m < 0
        ! 2) real a_lm =  complex a_lm                if m = 0
        ! 3) real a_lm =  sqrt(2) * RE{complex a_lm}  if m > 0
        ! real a_lms are stored in the (l**2 + l + m + 1)th 
        ! array element
        a_lm_real(:) = 0.0_dp
        do l = 0, l_max
            
            do m = -l, l
                
                index = l**2 + l + m! + 1 - 1
                if (m .lt. 0) then
                    
                    ! nb: may be losing precision here, as i'm not 
                    !     sure if aimag can handle double complexes
                    a_lm_real(index) = -sqrt(2.0_dp) * &
                                        aimag(a_lm_complex(1, l, -m))
                    
                else if(m .eq. 0) then
                    
                    a_lm_real(index) = real(a_lm_complex(1, l, m), &
                                            kind = dp)
                    
                else
                    
                    a_lm_real(index) = sqrt(2.0_dp) * &
                                       real(a_lm_complex(1, l, m), &
                                            kind = dp)
                    
                end if
                
            end do
            
        end do
        call cpu_time(time_1)
        time_elapsed = time_1 - time_0
        if (feedback) print*, 'a_lm conversion: completed in ', &
                              time_elapsed, ' secs'
        
        
    end subroutine a_lm_complex_to_real
    
    
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    
    
    ! draw random complex a_lms from specified power spectrum, 
    ! requiring an optional random-number-generator handle to be 
    ! specified. this is good for multiple calls to the same function.
    ! the RNG should be initialised with a seed in line with HEALPix 
    ! SYNFAST routine (see $HEALPIX/src/f90/synfast/syn_sub_inc.f90):
    ! "Seed should initially be set to a large, odd integer"
    subroutine c_l_to_a_lm(c_l, a_lm_complex, rng_handle)
        
        real(kind = dp), dimension(0:), intent(in) :: c_l
        complex(kind = dpc), dimension(1:, 0:, 0:), intent(inout) :: &
             a_lm_complex
        type(planck_rng), intent(inout) :: rng_handle
        real(kind = dp), dimension(:), allocatable :: a_lm_real
        real(kind = dp) :: sigma
        integer :: l_max, m_max, l, m, index, n_real_a_lms, seed
        
        
        ! determine useful params and allocate real a_lm array
        ! note that there are l_max + 1 C_ls in the array provided
        ! (0 => l_max, inclusive), so l_max is size(c_l) - 1
        l_max = size(c_l, dim = 1) - 1
        m_max = l_max
        n_real_a_lms = (l_max + 1) ** 2
        allocate(a_lm_real(0: n_real_a_lms - 1))
        
        
        ! draw random real a_lms and convert to complex
        do l = 0, l_max
            
            sigma = sqrt(c_l(l))
            do m = -l, l
                
                index = l**2 + l + m! + 1 - 1
                a_lm_real(index) = sigma * rand_gauss(rng_handle)
                
            end do
            
        end do
        call a_lm_real_to_complex(a_lm_real, a_lm_complex)
        deallocate(a_lm_real)
        
        
    end subroutine c_l_to_a_lm
    
    
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    
    
end program inpainting_test
